/* $XConsortium: copyright.c,v 1.2 94/11/21 21:55:13 kaleb Exp $ */
/* $XFree86: xc/programs/Xserver/hw/xfree86/reconfig/copyright.c,v 3.0 1994/09/21 10:55:51 dawes Exp $ */

/* Print the copyright to stdout */

copyright()
{
  printf("# Copyright (c) 1994 by The XFree86 Project, Inc.\n");

  printf("#\n");

  printf("# Permission is hereby granted, free of charge, to any person ");
  printf("obtaining a\n");

  printf("# copy of this software and associated documentation files (the ");
  printf("\"Software\"),\n");

  printf("# to deal in the Software without restriction, including without ");
  printf("limitation\n");

  printf("# the rights to use, copy, modify, merge, publish, distribute, ");
  printf("sublicense,\n");

  printf("# and/or sell copies of the Software, and to permit persons to ");
  printf("whom the\n");

  printf("# Software is furnished to do so, subject to the following ");
  printf("conditions:\n");

  printf("#\n");

  printf("# The above copyright notice and this permission notice shall be ");
  printf("included in\n");

  printf("# all copies or substantial portions of the Software.\n");

  printf("#\n");

  printf("# THE SOFTWARE IS PROVIDED \"AS IS\", WITHOUT WARRANTY OF ANY ");
  printf("KIND, EXPRESS OR\n");

  printf("# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF ");
  printf("MERCHANTABILITY,\n");

  printf("# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO ");
  printf("EVENT SHALL\n");

  printf("# THE XFREE86 PROJECT BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER ");
  printf("LIABILITY,\n");

  printf("# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING ");
  printf("FROM, OUT OF\n");

  printf("# OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS ");
  printf("IN THE\n");

  printf("# SOFTWARE.\n");

  printf("#\n");

  printf("# Except as contained in this notice, the name of the XFree86 ");
  printf("Project shall\n");

  printf("# not be used in advertising or otherwise to promote the sale, ");
  printf("use or other\n");

  printf("# dealings in this Software without prior written authorization ");
  printf("from the\n");

  printf("# XFree86 Project.\n");

  printf("#\n\n");
}
