/* $XConsortium: ppc.h,v 1.1 94/03/28 21:35:27 dpw Exp $ */
/*
 * Copyright IBM Corporation 1987,1988,1989
 *
 * All Rights Reserved
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appear in all copies and that 
 * both that copyright notice and this permission notice appear in
 * supporting documentation, and that the name of IBM not be
 * used in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.
 *
 * IBM DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
 * ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL
 * IBM BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR
 * ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
 * ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
 * SOFTWARE.
 *
*/
/* Header: /andrew/X11/r3src/release/server/ddx/ibm/ppc/RCS/ppc.h,v 9.2 89/04/29 21:25:30 jeff Exp */
/* Source: /andrew/X11/r3src/release/server/ddx/ibm/ppc/RCS/ppc.h,v */
/* "@(#)ppc.h	3.1 88/09/22 09:35:08" */

#include "gcstruct.h"

#include "ppcGCstr.h"

#include "ppcScrn.h"
#include "vga16.h"
